/**
 * @file ntbasic.c
 * @author Shinichiro Nakamura
 */

/*
 * ===============================================================
 * "Natural Tiny Basic (NT-Basic)"
 * "A tiny BASIC interpreter"
 * ---------------------------------------------------------------
 * NT-Basic interface module
 * ===============================================================
 * Copyright (c) 2012 Shinichiro Nakamura
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * ===============================================================
 */

#include "ntlibc.h"
#include "ntbasic.h"
#include "statement.h"
#include "expression.h"
#include "core.h"
#include "hal.h"

#define ERROR_RETURN(E) \
  do { \
    ntbasic_error_t error_number = (E); \
    if (error_number != NoError) { \
      return error_number; \
    } \
  } while (0)

/* Assign a variable a value. */
static ntbasic_error_t assignment(core_t *core)
{
  int value;
  char varname[VARIABLE_NAMESIZE];

  /* get the variable name */
  ERROR_RETURN(core_get_token(core));
  if (!ntlibc_isalpha(*core->token.text)) {
    return NotAVariable;
  }
  ntlibc_strcpy(varname, core->token.text);

  /* get the equals sign */
  ERROR_RETURN(core_get_token(core));
  if (*core->token.text != '=') {
    return EqualsSignExpected;
  }

  /* get the value to assign to var */
  ERROR_RETURN(expression_entry(core, &value));

  /* assign the value */
  if (variable_set(&(core->variable), varname, value) != 0) {
    return VariableAssignment;
  }

  return NoError;
}

ntbasic_error_t ntbasic_setup(core_t *core, char *program)
{
  hal_init();
  return core_setup(core, program);
}

ntbasic_error_t ntbasic_execute(core_t *core)
{
  ERROR_RETURN(core_get_token(core));
  /* check for assignment statement */
  if(core->token.type == VARIABLE) {
    core_putback(core); /* return the var to the input stream */
    ERROR_RETURN(assignment(core)); /* must be assignment statement */
  } else {
    /* is command */
    switch (core->token.statement) {
      case PRINT:
        ERROR_RETURN(statement_print(core));
        break;
      case GOTO:
        ERROR_RETURN(statement_goto(core));
        break;
      case IF:
        ERROR_RETURN(statement_if(core));
        break;
      case FOR:
        ERROR_RETURN(statement_for(core));
        break;
      case NEXT:
        ERROR_RETURN(statement_next(core));
        break;
      case INPUT:
        ERROR_RETURN(statement_input(core));
        break;
      case GOSUB:
        ERROR_RETURN(statement_gosub(core));
        break;
      case RETURN:
        ERROR_RETURN(statement_return(core));
        break;
      case REM:
        ERROR_RETURN(statement_rem(core));
        break;
      case READ:
        ERROR_RETURN(statement_read(core));
        break;
      case WRITE:
        ERROR_RETURN(statement_write(core));
        break;
      case END:
        // Nothing to do.
        break;
      case STNONE:
        // Nothing to do.
        break;
      case THEN:
        // Nothing to do.
        break;
      case TO:
        // Nothing to do.
        break;
      case EOL:
        // Nothing to do.
        break;
      case FINISHED:
        // Nothing to do.
        break;
    }
  }
  if ((core->token.statement == FINISHED) || (core->token.statement == END)) {
    return EndOfTheProgram;
  }
  return NoError;
}

