/**
 * @file statement.c
 * @author Shinichiro Nakamura
 */

/*
 * ===============================================================
 * "Natural Tiny Basic (NT-Basic)"
 * "A tiny BASIC interpreter"
 * ---------------------------------------------------------------
 * Core statement module
 * ===============================================================
 * Copyright (c) 2012 Shinichiro Nakamura
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * ===============================================================
 */

#include "ntlibc.h"
#include "statement.h"
#include "expression.h"
#include "core.h"
#include "hal.h"

#define ERROR_RETURN(E) \
  do { \
    ntbasic_error_t error_number = (E); \
    if (error_number != NoError) { \
      return error_number; \
    } \
  } while (0)

/* Push function for the FOR stack. */
static ntbasic_error_t for_stack_push(core_t *core, for_stack_t i)
{
  if (core->ftos >= FOR_MAX_DEPTH) {
    return TooManyNestedFORLoops;
  }

  core->fstack[core->ftos] = i;
  core->ftos++;
  return NoError;
}

static ntbasic_error_t for_stack_pop(core_t *core, for_stack_t *stack)
{
  core->ftos--;
  if (core->ftos < 0) {
    return NEXTWithoutFOR;
  }
  *stack = core->fstack[core->ftos];
  return NoError;
}

/* GOSUB stack push function. */
static ntbasic_error_t gosub_push(core_t *core, gosub_stack_t i)
{
  if (core->gtos >= GOSUB_MAX_DEPTH) {
    return TooManyNestedGOSUBs;
  }

  core->gtos++;
  core->gstack[core->gtos] = i;
  return NoError;
}

/* GOSUB stack pop function. */
static ntbasic_error_t gosub_pop(core_t *core, gosub_stack_t *stack)
{
  if (core->gtos == 0) {
    return RETURNWithoutGOSUB;
  }

  *stack = core->gstack[core->gtos--];
  return NoError;
}

static int print_integer(int number)
{
  char buf[16];
  ntlibc_puts(ntlibc_utoa(number, buf, 10));
  return ntlibc_strlen(buf);
}

/**
 * @brief Execute a simple version of the BASIC PRINT statement
 *
 * @param core core object.
 */
ntbasic_error_t statement_print(core_t *core)
{
  int answer;
  int len = 0, spaces;
  char last_delim = 0;

  do {
    ERROR_RETURN(core_get_token(core)); /* get next list item */
    if ((core->token.statement == EOL) || (core->token.statement == FINISHED)) {
      break;
    }
    if (core->token.type == QUOTE) {
      /*
       * is string
       */
      ntlibc_puts(core->token.text);
      len += ntlibc_strlen(core->token.text);
      ERROR_RETURN(core_get_token(core));
    } else {
      /*
       * is expression
       */
      core_putback(core);
      ERROR_RETURN(expression_entry(core, &answer));
      ERROR_RETURN(core_get_token(core));
      len += print_integer(answer);
    }
    last_delim = *core->token.text;

    if (*core->token.text == ';') {
      /*
       * compute number of spaces to move to next tab
       */
      spaces = 8 - (len % 8);
      len += spaces; /* add in the tabbing position */
      while (spaces) {
        ntlibc_puts(" ");
        spaces--;
      }
    } else if (*core->token.text == ',') {
      /*
       * do nothing
       */
    } else if ((core->token.statement != EOL) && (core->token.statement != FINISHED)) {
      return SyntaxError;
    }
  } while ((*core->token.text == ';') || (*core->token.text == ','));

  /*
   * Check the current statement.
   */
  if ((core->token.statement == EOL) || (core->token.statement == FINISHED)) {
    if ((last_delim != ';') && (last_delim != ',')) {
      ntlibc_puts("\n");
    }
  } else {
    return SyntaxError; /* error is not , or ; */
  }
  return NoError;
}

/**
 * @brief Execute a GOTO statement.
 */
ntbasic_error_t statement_goto(core_t *core)
{
  char *loc;

  ERROR_RETURN(core_get_token(core)); /* get label to go to */
  /* find the location of the label */
  loc = core_find_label(core, core->token.text);
  if (loc == '\0') {
    return UndefinedLabel; /* label not defined */
  } else {
    core->location.pointer = loc;  /* start core running at that loc */
  }
  return NoError;
}

/* Execute an IF statement. */
ntbasic_error_t statement_if(core_t *core)
{
  int x , y, cond;
  char op;

  expression_entry(core, &x); /* get left expression */

  ERROR_RETURN(core_get_token(core)); /* get the operator */
  if (!ntlibc_strchr("=<>", *core->token.text)) {
    return SyntaxError; /* not a legal operator */
  }
  op = *core->token.text;

  expression_entry(core, &y); /* get right expression */

  /* determine the outcome */
  cond = 0;
  switch (op) {
    case '<':
      if (x < y) {
        cond = 1;
      }
      break;
    case '>':
      if (x > y) {
        cond = 1;
      }
      break;
    case '=':
      if (x == y) {
        cond = 1;
      }
      break;
  }
  if (cond) { /* is true so process target of IF */
    ERROR_RETURN(core_get_token(core));
    if (core->token.statement != THEN) {
      return THENExpected;
    }/* else core execution starts on next line */
  } else {
    core_find_eol(core); /* find start of next line */
  }
  return NoError;
}

/* Execute a FOR loop. */
ntbasic_error_t statement_for(core_t *core)
{
  for_stack_t i;
  int value;
  int n;

  /* read the control variable */
  ERROR_RETURN(core_get_token(core));
  if (!ntlibc_isalpha(*core->token.text)) {
    return NotAVariable;
  }
  ntlibc_strcpy(i.varname, core->token.text);

  ERROR_RETURN(core_get_token(core)); /* read the equals sign */
  if (*core->token.text != '=') {
    return EqualsSignExpected;
  }

  /* get initial value */
  expression_entry(core, &value);
  if (variable_set(&(core->variable), i.varname, value) != 0) {
    return VariableAssignment;
  }

  ERROR_RETURN(core_get_token(core));
  if (core->token.statement != TO) {
    return TOExpected; /* read and discard the TO */
  }

  expression_entry(core, &i.target); /* get target value */

  if (variable_get(&(core->variable), i.varname, &n) != 0) {
    return VariableAssignment;
  }

  /* if loop can execute at least once, push info on stack */
  if (value >= n) {
    i.loc = core->location.pointer;
    ERROR_RETURN(for_stack_push(core, i));
  } else {
    /* otherwise, skip loop code altogether */
    while (core->token.statement != NEXT) {
      ERROR_RETURN(core_get_token(core));
    }
  }
  return NoError;
}

/* Execute a NEXT statement. */
ntbasic_error_t statement_next(core_t *core)
{
  for_stack_t i;
  int n;

  ERROR_RETURN(for_stack_pop(core, &i)); /* read the loop info */

  if (variable_get(&(core->variable), i.varname, &n) != 0) {
    return VariableAssignment;
  }
  n++;
  if (variable_set(&(core->variable), i.varname, n) != 0) {
    return VariableAssignment;
  }
  if (n > i.target) {
    return NoError;  /* all done */
  }
  ERROR_RETURN(for_stack_push(core, i));  /* otherwise, restore the info */
  core->location.pointer = i.loc;  /* loop */
  return NoError;
}

/* Execute a simple form of the BASIC INPUT command */
ntbasic_error_t statement_input(core_t *core)
{
  char buf[80];
  char varname[VARIABLE_NAMESIZE];
  int n;

  ERROR_RETURN(core_get_token(core)); /* see if prompt string is present */
  if (core->token.type == QUOTE) {
    ntlibc_puts(core->token.text); /* if so, print it and check for comma */
    ERROR_RETURN(core_get_token(core));
    if (*core->token.text != ',') {
      return UnbalancedParentheses;
    }
    ERROR_RETURN(core_get_token(core));
  } else {
    ntlibc_puts("? "); /* otherwise, prompt with / */
  }
  ntlibc_strcpy(varname, core->token.text);

  n = ntlibc_atoi(ntlibc_gets(buf, sizeof(buf)));    /* read input */

  if (variable_set(&(core->variable), varname, n) != 0) {
    return VariableAssignment;
  }
  return NoError;
}

/* Execute a GOSUB command. */
ntbasic_error_t statement_gosub(core_t *core)
{
  gosub_stack_t i;
  char *loc;

  ERROR_RETURN(core_get_token(core));
  /* find the label to call */
  loc = core_find_label(core, core->token.text);
  if (loc == '\0') {
    return UndefinedLabel; /* label not defined */
  } else {
    i.loc = core->location.pointer;
    ERROR_RETURN(gosub_push(core, i)); /* save place to return to */
    core->location.pointer = loc;  /* start core running at that loc */
  }
  return NoError;
}

/* Return from GOSUB. */
ntbasic_error_t statement_return(core_t *core)
{
  gosub_stack_t i;
  ERROR_RETURN(gosub_pop(core, &i));
  core->location.pointer = i.loc;
  return NoError;
}

ntbasic_error_t statement_rem(core_t *core)
{
  core_find_eol(core);   /* find start of next line */
  return NoError;
}

ntbasic_error_t statement_read(core_t *core)
{
  unsigned int port, value;
  char varname[VARIABLE_NAMESIZE];
  int n;

  ERROR_RETURN(core_get_token(core));
  if (core->token.type != NUMBER) {
    ntlibc_strcpy(varname, core->token.text);
    if (variable_get(&(core->variable), varname, &n) != 0) {
      return VariableAssignment;
    }
    port = n;
  } else {
    port = ntlibc_atoi(core->token.text);
  }

  ERROR_RETURN(core_get_token(core));
  if (*core->token.text != ',') {
    return SyntaxError;
  }

  ERROR_RETURN(core_get_token(core));
  ntlibc_strcpy(varname, core->token.text);

  hal_read(port, &value);
  if (variable_set(&(core->variable), varname, value) != 0) {
    return VariableAssignment;
  }

  core_find_eol(core); /* find start of next line */
  return NoError;
}

/**
 * @brief Write to the hardware abstract layer.
 * @details WRITE <port>, <value>
 */
ntbasic_error_t statement_write(core_t *core)
{
  int port, value;
  char varname[VARIABLE_NAMESIZE];
  int n;

  ERROR_RETURN(core_get_token(core));
  if (core->token.type != NUMBER) {
    ntlibc_strcpy(varname, core->token.text);
    if (variable_get(&(core->variable), varname, &n) != 0) {
      return VariableAssignment;
    }
    port = n;
  } else {
    port = ntlibc_atoi(core->token.text);
  }

  ERROR_RETURN(core_get_token(core));
  if (*core->token.text != ',') {
    return SyntaxError;
  }

  ERROR_RETURN(core_get_token(core));
  if (core->token.type != NUMBER) {
    ntlibc_strcpy(varname, core->token.text);
    if (variable_get(&(core->variable), varname, &n) != 0) {
      return VariableAssignment;
    }
    value = n;
  } else {
    value = ntlibc_atoi(core->token.text);
  }

  hal_write(port, value);
  core_find_eol(core); /* find start of next line */
  return NoError;
}

