/**
 * @file variable.c
 * @author Shinichiro Nakamura
 */

/*
 * ===============================================================
 * "Natural Tiny Basic (NT-Basic)"
 * "A tiny BASIC interpreter"
 * ---------------------------------------------------------------
 * Variable module
 * ===============================================================
 * Copyright (c) 2012 Shinichiro Nakamura
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * ===============================================================
 */

#include "variable.h"
#include "ntlibc.h"

static unsigned int hashed_index(const char *name)
{
  unsigned int n = 0;
  char *p = (char *)name;
  while (*p) {
    n += *p++;
  }
  return n % VARIABLE_LISTSIZE;
}

int variable_init(variable_t *p)
{
  int i;
  for (i = 0; i < VARIABLE_LISTSIZE; i++) {
    *p->symbols[i].name = '\0';
    p->symbols[i].value = 0;
  }
  return 0;
}

int variable_set(variable_t *p, const char *name, const int value)
{
  const int index = hashed_index(name);
  if (ntlibc_strlen(name) >= VARIABLE_NAMESIZE) {
    return -1;
  }
  if (ntlibc_strcmp(p->symbols[index].name, name) == 0) {
    p->symbols[index].value = value;
    return 0;
  } else {
    int i;
    /*
     * Search the exists entry.
     */
    for (i = 0; i < VARIABLE_LISTSIZE; i++) {
      int n = (index + i) % VARIABLE_LISTSIZE;
      if (ntlibc_strcmp(p->symbols[n].name, name) == 0) {
        p->symbols[n].value = value;
        return 0;
      }
    }
    /*
     * Create the new entry.
     */
    for (i = 0; i < VARIABLE_LISTSIZE; i++) {
      int n = (index + i) % VARIABLE_LISTSIZE;
      if (ntlibc_strlen(p->symbols[n].name) == 0) {
        ntlibc_strcpy(p->symbols[n].name, name);
        p->symbols[n].value = value;
        return 0;
      }
    }
  }
  ntlibc_puts("[variable_set: Could not entry the variable:");
  ntlibc_puts(name);
  ntlibc_puts("]\n");
  return -1;
}

int variable_get(variable_t *p, const char *name, int *value)
{
  const int index = hashed_index(name);
  if (ntlibc_strlen(name) >= VARIABLE_NAMESIZE) {
    return -1;
  }
  if (ntlibc_strcmp(p->symbols[index].name, name) == 0) {
    *value = p->symbols[index].value;
    return 0;
  } else {
    int i;
    /*
     * Search the exists entry.
     */
    for (i = 0; i < VARIABLE_LISTSIZE; i++) {
      int n = (index + i) % VARIABLE_LISTSIZE;
      if (ntlibc_strcmp(p->symbols[n].name, name) == 0) {
        *value = p->symbols[n].value;
        return 0;
      }
    }
    /*
     * Create the new entry.
     */
    if (variable_set(p, name, 0) == 0) {
      *value = 0;
      return 0;
    }
  }
  ntlibc_puts("[variable_get: Not found the variable:");
  ntlibc_puts(name);
  ntlibc_puts("]\n");
  *value = 0;
  return -1;
}

