/**
 * @file textedit.c
 * @author Shinichiro Nakamura
 */

/*
 * ===============================================================
 * "Natural Tiny Basic (NT-Basic)"
 * "A tiny BASIC interpreter"
 * ---------------------------------------------------------------
 * Text edit module
 * ===============================================================
 * Copyright (c) 2012 Shinichiro Nakamura
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * ===============================================================
 */

#include "textedit.h"
#include "ntlibc.h"

#define INITCODE    (0x14376329)

static int is_delim(int c)
{
  return ((c == '\n') || (c == '\0')) ? 1 : 0;
}

static char *get_linenext(char *src)
{
  while (!is_delim(*src)) {
    src++;
  }
  if (*src == '\0') {
    return 0;
  } else {
    return ++src;
  }
}

static char *get_linetext(char *src, char *des, int siz)
{
  int cnt = 0;
  while (!is_delim(*src)) {
    *des = *src;
    des++;
    src++;
    cnt++;
    if (siz - 1 <= cnt) {
      break;
    }
  }
  *des = '\0';
  if (*src == '\0') {
    return 0;
  } else {
    return ++src;
  }
}

static int get_linelength(char *src)
{
  int cnt = 0;
  while (!is_delim(*src)) {
    src++;
    cnt++;
  }
  return cnt;
}

int textedit_init(textedit_t *p)
{
  unsigned int i;
  p->initcode = INITCODE;
  for (i = 0; i < sizeof(p->textdata); i++) {
    p->textdata[i] = 0;
  }
  return 0;
}

int textedit_clear(textedit_t *p)
{
  unsigned int i;
  if (p->initcode != INITCODE) {
    return -1;
  }
  for (i = 0; i < sizeof(p->textdata); i++) {
    p->textdata[i] = 0;
  }
  return 0;
}

int textedit_insert(textedit_t *p, const int line_number, const char *text)
{
  char *s = 0;
  int cnt = 0;

  if (p->initcode != INITCODE) {
    return -1;
  }

  s = p->textdata;
  do {
    if (cnt == line_number) {
      char *src, *des;
      /*
       * 1st step.
       * Create a blank for the new text string.
       */
      src = s + ntlibc_strlen(s);
      des = s + ntlibc_strlen(s) + ntlibc_strlen(text);
      do {
        *des = *src;
        des--;
        src--;
      } while (s <= src);
      /*
       * 2nd step.
       * Insert the new text string.
       */
      src = (char *)text;
      des = s;
      while (*src) {
        *des = *src;
        des++;
        src++;
      }
      return 0;
    }
    s = get_linenext(s);
    cnt++;
  } while (s != 0);

  if (cnt == line_number) {
    char *src = (char *)text;
    char *des = p->textdata + ntlibc_strlen(p->textdata);
    while (*src) {
      *des = *src;
      des++;
      src++;
    }
    *des = '\0';
    return 0;
  }

  return -1;
}

int textedit_delete(textedit_t *p, const int line_number)
{
  char *s = 0;
  int cnt = 0;

  if (p->initcode != INITCODE) {
    return -1;
  }

  s = p->textdata;
  do {
    if (cnt == line_number) {
      char *src, *des;
      src = s + get_linelength(s) + 1;
      des = s;
      if (p->textdata + ntlibc_strlen(p->textdata) <= src) {
        *des = '\0';
        return 0;
      } else {
        while (*src) {
          *des = *src;
          des++;
          src++;
        }
        *des = '\0';
        return 0;
      }
    }
    s = get_linenext(s);
    cnt++;
  } while (s != 0);

  return -1;
}

int textedit_fetch(textedit_t *p, const int line_number, char *text, const int maxsize)
{
  char *s = 0;
  int cnt = 0;

  if (p->initcode != INITCODE) {
    return -1;
  }

  s = p->textdata;
  do {
    s = get_linetext(s, text, maxsize);
    if (cnt == line_number) {
      if (s == 0) {
        return -1;
      } else {
        return 0;
      }
    } else {
      *text = '\0';
    }
    cnt++;
  } while (s != 0);

  return -1;
}

int textedit_info(textedit_t *p, textedit_info_t *info)
{
  unsigned int i;
  unsigned int lines = 0;
  if (p->initcode != INITCODE) {
    return -1;
  }
  for (i = 0; i < sizeof(p->textdata); i++) {
    if (p->textdata[i] == '\n') {
      lines++;
    }
    if (p->textdata[i] == '\0') {
      break;
    }
  }
  info->lines = lines;
  return 0;
}

